import unittest
import multiprocessing
from molotov.result import LiveResults
from molotov.tests.support import async_test


class LiveResultsTest(unittest.TestCase):

    @async_test
    async def test_live(self, loop):

        # we have results
        live = LiveResults(use_buffer=False)

        def _do_it():
            live.incr_success()
            live.incr_failure()

        # we fork 2 processes
        p1 = multiprocessing.Process(target=_do_it)
        p2 = multiprocessing.Process(target=_do_it)
        p1.start()
        p2.start()
        p1.join()
        p2.join()
        pids = [p1.pid, p2.pid]
        live.update()
        self.assertEqual(live.get_successes(), 2)
        self.assertEqual(live.get_failures(), 2)
        for pid in pids:
            self.assertTrue(live.get_successes(pid) > 0)
            self.assertTrue(live.get_failures(pid) > 0)



import socket
from urllib.parse import urlparse
import asyncio
from aiohttp.client import ClientSession, ClientRequest
from aiohttp import TCPConnector
from molotov.util import resolve


_HOST = socket.gethostname()
_UNREADABLE = "***WARNING: Molotov can't display this body***"
_BINARY = "**** Binary content ****"
_COMPRESSED = ('gzip', 'compress', 'deflate', 'identity', 'br')


class LoggedClientRequest(ClientRequest):
    session = None

    def send(self, writer, reader):
        if self.session and self.verbose:
            info = self.session.print_request(self)
            asyncio.ensure_future(info)
        return super(LoggedClientRequest, self).send(writer, reader)


class LoggedClientSession(ClientSession):

    def __init__(self, loop, stream, verbose=False, statsd=None, **kw):
        connector = kw.pop('connector', None)
        if connector is None:
            connector = TCPConnector(loop=loop, limit=None)
        super(LoggedClientSession,
              self).__init__(loop=loop, request_class=LoggedClientRequest,
                             connector=connector,  **kw)
        self.stream = stream
        self.request_class = LoggedClientRequest
        self.request_class.verbose = verbose
        self.verbose = verbose
        self.request_class.session = self
        self.statsd = statsd

    def _dns_lookup(self, url):
        return resolve(url)[0]

    async def _request(self, *args, **kw):
        args = list(args)
        args[1] = self._dns_lookup(args[1])
        args = tuple(args)
        req = super(LoggedClientSession, self)._request

        if self.statsd:
            prefix = 'molotov.%(hostname)s.%(method)s.%(host)s.%(path)s'
            meth, url = args[:2]
            url = urlparse(url)
            path = url.path != '' and url.path or '/'

            data = {'method': meth,
                    'hostname': _HOST,
                    'host': url.netloc.split(":")[0],
                    'path': path}

            label = prefix % data

            @self.statsd.timer(label)
            async def request():
                resp = await req(*args, **kw)
                self.statsd.incr(label + '.' + str(resp.status))
                return resp

            resp = await request()
        else:
            resp = await req(*args, **kw)

        await self.print_response(resp)
        return resp

    async def print_request(self, req):
        if not self.verbose:
            return

        await self.stream.put('>' * 45)
        raw = '\n' + req.method + ' ' + str(req.url)
        if len(req.headers) > 0:
            headers = '\n'.join('%s: %s' % (k, v) for k, v in
                                req.headers.items())
            raw += '\n' + headers

        if req.headers.get('Content-Encoding') in _COMPRESSED:
            raw += '\n\n' + _BINARY + '\n'
        elif req.body:
            if isinstance(req.body, bytes):
                try:
                    body = str(req.body, 'utf8')
                except UnicodeDecodeError:
                    body = _UNREADABLE
            else:
                body = req.body

            raw += '\n\n' + body + '\n'
        await self.stream.put(raw)

    async def print_response(self, resp):
        if not self.verbose:
            return
        await self.stream.put('\n' + '=' * 45 + '\n')
        raw = 'HTTP/1.1 %d %s\n' % (resp.status, resp.reason)
        items = resp.headers.items()
        headers = '\n'.join('{}: {}'.format(k, v) for k, v in items)
        raw += headers
        if resp.headers.get('Content-Encoding') in _COMPRESSED:
            raw += '\n\n' + _BINARY
        elif resp.content:
            content = await resp.content.read()
            # put back the data in the content
            resp.content.unread_data(content)
            try:
                raw += '\n\n' + content.decode()
            except UnicodeDecodeError:
                raw += '\n\n' + _UNREADABLE

        await self.stream.put(raw)
        await self.stream.put('\n' + '<' * 45 + '\n')



import asyncio
import unittest
import sys
from molotov.slave import main


class TestSlave(unittest.TestCase):
    def test_main(self):
        oldloop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        saved = list(sys.argv)
        repo = 'https://github.com/loads/molotov'
        run = 'test'
        sys.argv[:] = ['moloslave', repo, run]
        try:
            main()
        finally:
            sys.argv[:] = saved
            asyncio.set_event_loop(oldloop)



try:
    from aiostatsd.client import StatsdClient
except ImportError:
    StatsdClient = None


def get_statsd_client(server="127.0.0.1", port=8125, packet_size=512,
                      flush_interval=.01):
    if StatsdClient is None:
        raise ImportError("You need to install aiostatsd")
    client = StatsdClient(server, port, packet_size=packet_size,
                          flush_interval=flush_interval)
    return client



""" Molotov-based test.
"""
import json
from molotov import scenario, setup, global_setup, teardown, global_teardown


# This is the service you want to load test
_API = 'http://localhost:8080'


@global_setup
def test_starts(args):
    """ This functions is called before anything starts.

    Notice that it's not a coroutine.
    """
    pass


@setup()
async def worker_starts(worker_id, args):
    """ This function is called once per worker.

    If it returns a mapping, it will be used with all requests.

    You can add things like Authorization headers for instance,
    by setting a "headers" key.
    """
    headers = {'SomeHeader': '1'}
    return {'headers': headers}


@teardown
def worker_ends(worker_id):
    """ This functions is called when the worker is done.

    Notice that it's not a coroutine.
    """
    pass


@global_teardown
def test_ends():
    """ This functions is called when everything is done.

    Notice that it's not a coroutine.
    """
    pass


# each scenario has a weight. Molotov uses it to determine
# how often the scenario is picked.
@scenario(40)
async def scenario_one(session):
    async with session.get(_API) as resp:
        # if Molotov is called with --statsd
        # you will have a statsd client set into the session
        # you can use to add metrics
        if session.statsd:
            session.statsd.incr('BLEH')
        # when you read the body, don't forget to use await
        res = await resp.json()
        assert res['result'] == 'OK'
        assert resp.status == 200


# all scenarii are coroutines
@scenario(30)
async def scenario_two(session):
    # a call to one of the session method should be awaited
    # see aiohttp.Client docs for more info on this
    async with session.get(_API) as resp:
        assert resp.status == 200


@scenario(30)
async def scenario_three(session):
    somedata = json.dumps({'OK': 1})
    async with session.post(_API, data=somedata) as resp:
        assert resp.status == 200



import json
from molotov import scenario, setup, global_setup, global_teardown, teardown


_API = 'http://localhost:8080'
_HEADERS = {}


# notice that the global setup, global teardown and teardown
# are not a coroutine.
@global_setup()
def init_test(args):
    _HEADERS['SomeHeader'] = '1'


@global_teardown()
def end_test():
    print("This is the end")


@setup()
async def init_worker(worker_num, args):
    headers = {'AnotherHeader': '1'}
    headers.update(_HEADERS)
    return {'headers': headers}


@teardown()
def end_worker(worker_num):
    print("This is the end for %d" % worker_num)


@scenario(40)
async def scenario_one(session):
    async with session.get(_API) as resp:
        if session.statsd:
            session.statsd.incr('BLEH')
        res = await resp.json()
        assert res['result'] == 'OK'
        assert resp.status == 200


@scenario(30)
async def scenario_two(session):
    async with session.get(_API) as resp:
        assert resp.status == 200


@scenario(30)
async def scenario_three(session):
    somedata = json.dumps({'OK': 1})
    async with session.post(_API, data=somedata) as resp:
        assert resp.status == 200



import sys
import signal
import os
import asyncio
import unittest
import multiprocessing
import time
from contextlib import contextmanager
import functools
from collections import namedtuple
from http.client import HTTPConnection
from io import StringIO

from aiohttp.client_reqrep import ClientResponse, URL
from multidict import CIMultiDict
from molotov.api import _SCENARIO, _FIXTURES
from molotov import fmwk


def run_server(port=8888):
    """Running in a subprocess to avoid any interference
    """
    def _run():
        import http.server
        import socketserver
        import signal
        import sys

        socketserver.TCPServer.allow_reuse_address = True
        Handler = http.server.SimpleHTTPRequestHandler
        attempts = 0
        httpd = None

        while attempts < 3:
            try:
                httpd = socketserver.TCPServer(("", port), Handler)
                break
            except Exception:
                attempts += 1
                time.sleep(.1)

        if httpd is None:
            raise OSError("Could not start the coserver")

        def _shutdown(*args, **kw):
            httpd.server_close()
            sys.exit(0)

        signal.signal(signal.SIGTERM, _shutdown)
        signal.signal(signal.SIGINT, _shutdown)
        httpd.serve_forever()

    p = multiprocessing.Process(target=_run)
    p.start()
    start = time.time()
    connected = False

    while time.time() - start < 5 and not connected:
        try:
            conn = HTTPConnection('localhost', 8888)
            conn.request("GET", "/")
            conn.getresponse()
            connected = True
        except Exception:
            time.sleep(.1)
    if not connected:
        os.kill(p.pid, signal.SIGTERM)
        p.join(timeout=1.)
        raise OSError('Could not connect to coserver')
    return p


_CO = {'clients': 0, 'server': None}


@contextmanager
def coserver(port=8888):
    if _CO['clients'] == 0:
        _CO['server'] = run_server(port)

    _CO['clients'] += 1
    try:
        yield
    finally:
        _CO['clients'] -= 1
        if _CO['clients'] == 0:
            os.kill(_CO['server'].pid, signal.SIGTERM)
            _CO['server'].join(timeout=1.)
            _CO['server'] = None


def Response(method='GET', status=200, body=b'***'):
    response = ClientResponse(method, URL('/'))
    response.status = status
    response.reason = ''
    response.code = status
    response.should_close = False
    response.headers = CIMultiDict({})
    response.raw_headers = []

    class Body:
        async def read(self):
            return body

        def unread_data(self, data):
            pass

    response.content = Body()
    response._content = body

    return response


class TestLoop(unittest.TestCase):
    def setUp(self):
        self.old = list(_SCENARIO)
        self.oldsetup = dict(_FIXTURES)
        fmwk._STOP = False

    def tearDown(self):
        _SCENARIO[:] = self.old
        _FIXTURES.clear()
        _FIXTURES.update(self.oldsetup)

    def get_args(self):
        args = namedtuple('args', 'verbose quiet duration exception')
        args.verbose = True
        args.quiet = False
        args.duration = 1
        args.exception = True
        args.processes = 1
        args.debug = True
        args.workers = 1
        args.console = True
        args.statsd = False
        return args


def async_test(func):
    @functools.wraps(func)
    def _async_test(*args, **kw):
        cofunc = asyncio.coroutine(func)
        oldloop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_debug(True)
        kw['loop'] = loop
        try:
            loop.run_until_complete(cofunc(*args, **kw))
        finally:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(oldloop)
    return _async_test


def dedicatedloop(func):
    @functools.wraps(func)
    def _loop(*args, **kw):
        old_loop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return func(*args, **kw)
        finally:
            asyncio.set_event_loop(old_loop)
    return _loop


@contextmanager
def set_args(*args):
    old = list(sys.argv)
    sys.argv[:] = args
    oldout, olderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()
    try:
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout.seek(0)
        sys.stderr.seek(0)
        sys.argv[:] = old
        sys.stdout, sys.stderr = oldout, olderr



from functools import partial
import signal
import multiprocessing
import asyncio
import time
import sys
import os

from molotov.util import log, stream_log
from molotov.session import LoggedClientSession as Session
from molotov.result import LiveResults, ClosedError
from molotov.api import get_fixture, pick_scenario
from molotov.stats import get_statsd_client
from molotov.ui import quit as quit_screen


_STOP = False
_REFRESH = .3


def _now():
    return int(time.time())


_results = LiveResults()


def get_live_results():
    return _results


async def consume(queue, numworkers, console=False, verbose=False):
    worker_stopped = 0
    while True and worker_stopped < numworkers:
        try:
            item = await queue.get()
        except RuntimeError:
            break
        if item == 'WORKER_STOPPED':
            worker_stopped += 1
        elif item == 'STOP':
            break

        elif isinstance(item, str):
            results = get_live_results()
            try:
                if item == '.':
                    results.incr_success()
                elif item == '-':
                    results.incr_failure()
                else:
                    if console and verbose:
                        print(item)
            except ClosedError:
                break
        else:
            if console and verbose:
                import traceback
                traceback.print_tb(item)


async def step(session, quiet, verbose, stream):
    """ single scenario call.

    When it returns 1, it works. -1 the script failed,
    0 the test is stopping or needs to stop.
    """
    func, args_, kw = pick_scenario()
    try:
        await func(session, *args_, **kw)
        await stream.put('.')
        return 1
    except asyncio.CancelledError:
        return 0
    except Exception as exc:
        await stream.put('-')
        if verbose:
            await stream.put(repr(exc))
            await stream.put(sys.exc_info()[2])

    return -1


_HOWLONG = 0


async def worker(num, loop, results, args, stream, statsd):
    global _STOP
    quiet = args.quiet
    duration = args.duration
    verbose = args.verbose
    exception = args.exception
    count = 1
    start = _now()
    howlong = 0
    setup = get_fixture('setup')
    if setup is not None:
        try:
            options = await setup(num, args)
        except Exception as e:
            log(e)
            await stream.put('WORKER_STOPPED')
            return
        if options is None:
            options = {}
        elif not isinstance(options, dict):
            log('The setup function needs to return a dict')
            await stream.put('WORKER_STOPPED')
            return
    else:
        options = {}

    async with Session(loop, stream, verbose, statsd, **options) as session:
        while howlong < duration and not _STOP:
            howlong = _now() - start
            result = await step(session, quiet, verbose, stream)
            if result == 1:
                results['OK'] += 1
            elif result == -1:
                results['FAILED'] += 1
                if exception:
                    await stream.put('WORKER_STOPPED')
                    _STOP = True
            elif result == 0:
                break
            count += 1

    if not _STOP:
        await stream.put('WORKER_STOPPED')


def _worker_done(num, future):
    teardown = get_fixture('teardown')
    if teardown is not None:
        try:
            teardown(num)
        except Exception as e:
            # we can't stop the teardown process
            log(e)


def _runner(loop, args, results, stream, statsd):
    def _prepare():
        tasks = []
        for i in range(args.workers):
            future = asyncio.ensure_future(worker(i, loop, results, args,
                                                  stream, statsd))
            future.add_done_callback(partial(_worker_done, i))
            tasks.append(future)
        return tasks
    if args.quiet:
        return _prepare()
    else:
        with stream_log('Preparing %d workers' % args.workers):
            return _prepare()


def _process(args):
    global _STOP
    if args.processes > 1:
        signal.signal(signal.SIGINT, _shutdown)
        signal.signal(signal.SIGTERM, _shutdown)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    else:
        loop = asyncio.get_event_loop()

    if args.debug:
        log('**** RUNNING IN DEBUG MODE == SLOW ****')
        loop.set_debug(True)

    results = {'OK': 0, 'FAILED': 0}
    stream = asyncio.Queue(loop=loop)

    co_tasks = []

    if args.statsd:
        statsd = get_statsd_client(args.statsd_server, args.statsd_port)
        stastd_task = asyncio.ensure_future(statsd.run())
    else:
        statsd = stastd_task = None

    consumer = asyncio.ensure_future(consume(stream, args.workers,
                                     args.console, args.verbose))
    co_tasks.append(consumer)
    _TASKS.extend(co_tasks)

    co_tasks = asyncio.gather(*co_tasks, loop=loop, return_exceptions=True)

    workers = _runner(loop, args, results, stream, statsd)
    run_task = asyncio.gather(*workers, loop=loop, return_exceptions=True)

    _TASKS.extend(workers)
    _STATSD.append((statsd, stastd_task))

    try:
        loop.run_until_complete(run_task)
    except asyncio.CancelledError:
        _STOP = True
        co_tasks.cancel()
        loop.run_until_complete(co_tasks)
        run_task.cancel()
        loop.run_until_complete(run_task)
    finally:
        _stop_statsd()
        loop.close()

    return results


_STATSD = []
_PROCESSES = []
_TASKS = []


def _stop_statsd():

    if _STATSD != []:
        loop = asyncio.get_event_loop()

        async def stop():
            statsd, stastd_task = _STATSD[0]
            try:
                await statsd.stop()
                await stastd_task
            except Exception:
                pass

        stop = asyncio.ensure_future(stop())
        loop.run_until_complete(stop)
        _STATSD[:] = []


def _shutdown(signal, frame):
    global _STOP
    _STOP = True
    get_live_results().close()

    for task in _TASKS:
        task.cancel()

    # send sigterms
    for proc in _PROCESSES:
        proc.terminate()


def _launch_processes(args, screen):
    results = {'FAILED': 0, 'OK': 0}
    signal.signal(signal.SIGINT, _shutdown)
    signal.signal(signal.SIGTERM, _shutdown)

    if args.processes > 1:
        if not args.quiet:
            log('Forking %d processes' % args.processes, pid=False)
        result_queue = multiprocessing.Queue()
        ui = None
        loop = asyncio.get_event_loop()

        def _pprocess(result_queue):
            result_queue.put(_process(args))

        jobs = []
        for i in range(args.processes):
            p = multiprocessing.Process(target=_pprocess,
                                        args=(result_queue,))
            jobs.append(p)
            p.start()

        for job in jobs:
            _PROCESSES.append(job)

        if screen is not None and not args.console:
            if args.processes == 1:
                pids = [os.getpid()]
            else:
                pids = [job.pid for job in jobs]

            ui = screen(pids, get_live_results)

            def check_procs(*args):
                dead = [not p.is_alive() for p in _PROCESSES]
                if all(dead):
                    quit_screen()

            ui.set_alarm_in(1, check_procs)
            ui.run()

        async def run(loop, quiet, console):
            while len(_PROCESSES) > 0:
                if not quiet and console:
                    try:
                        print(get_live_results(), end='\r')
                    except ClosedError:
                        # finished
                        return
                for job in jobs:
                    if job.exitcode is not None and job in _PROCESSES:
                        _PROCESSES.remove(job)

                await asyncio.sleep(.2)

        loop.run_until_complete(asyncio.ensure_future(run(loop, args.quiet,
                                                          args.console)))

        for job in jobs:
            proc_result = result_queue.get()
            results['FAILED'] += proc_result['FAILED']
            results['OK'] += proc_result['OK']
    else:
        loop = asyncio.get_event_loop()
        if screen is not None and not args.console:
            ui = screen([os.getpid()], get_live_results, loop)
            ui.start()
        else:
            ui = None

        if not args.quiet and args.console and not args.verbose:
            def _display(loop):
                try:
                    print(get_live_results(), end='\r')
                except ClosedError:
                    return
                loop.call_later(_REFRESH, _display, loop)

            loop.call_soon(_display, loop)

        results = _process(args)

        if ui is not None:
            ui.stop()

    return results


def runner(args, screen=None):
    global_setup = get_fixture('global_setup')
    if global_setup is not None:
        global_setup(args)

    try:
        return _launch_processes(args, screen)
    finally:
        global_teardown = get_fixture('global_teardown')
        if global_teardown is not None:
            global_teardown()



import asyncio
import os
import signal

from molotov.session import LoggedClientSession
from molotov.fmwk import step, worker, runner
from molotov.api import (scenario, setup, global_setup, teardown,
                         global_teardown)
from molotov.tests.support import TestLoop, async_test


class TestFmwk(TestLoop):

    @async_test
    async def test_step(self, loop):
        res = []

        @scenario(0)
        async def test_one(session):
            res.append('1')

        @scenario(100)
        async def test_two(session):
            res.append('2')

        stream = asyncio.Queue()
        async with LoggedClientSession(loop, stream) as session:
            result = await step(session, False, False, stream)
            self.assertTrue(result, 1)
            self.assertEqual(len(res), 1)

    @async_test
    async def test_failing_step(self, loop):

        @scenario(100)
        async def test_two(session):
            raise ValueError()

        stream = asyncio.Queue()
        async with LoggedClientSession(loop, stream) as session:
            result = await step(session, False, False, stream)
            self.assertTrue(result, -1)

    @async_test
    async def test_aworker(self, loop):

        res = []

        @setup()
        async def setuptest(num, args):
            res.append('0')

        @scenario(50)
        async def test_one(session):
            pass

        @scenario(100)
        async def test_two(session):
            pass

        results = {'OK': 0, 'FAILED': 0}
        stream = asyncio.Queue()
        args = self.get_args()
        statsd = None

        await worker(1, loop, results, args, stream, statsd)

        self.assertTrue(results['OK'] > 0)
        self.assertEqual(results['FAILED'], 0)
        self.assertEqual(len(res), 1)

    def test_runner(self):
        res = []

        @global_setup()
        def init(args):
            res.append('SETUP')

        @setup()
        async def setuptest(num, args):
            res.append('0')

        @scenario(50)
        async def test_one(session):
            pass

        @scenario(100)
        async def test_two(session):
            pass

        args = self.get_args()
        results = runner(args)
        self.assertTrue(results['OK'] > 0)
        self.assertEqual(results['FAILED'], 0)
        self.assertEqual(len(res), 2)

    def test_runner_multiprocess(self):
        res = []

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        @setup()
        async def setuptest(num, args):
            res.append('0')

        @scenario(50)
        async def test_one(session):
            pass

        @scenario(100)
        async def test_two(session):
            pass

        args = self.get_args()
        args.processes = 2
        args.workers = 5
        results = runner(args)
        self.assertTrue(results['OK'] > 0)
        self.assertEqual(results['FAILED'], 0)

    @async_test
    async def test_aworker_noexc(self, loop):

        res = []

        @setup()
        async def setuptest(num, args):
            res.append('0')

        @scenario(50)
        async def test_one(session):
            pass

        @scenario(100)
        async def test_two(session):
            pass

        results = {'OK': 0, 'FAILED': 0}
        stream = asyncio.Queue()
        args = self.get_args()
        args.exception = False
        statsd = None

        await worker(1, loop, results, args, stream, statsd)

        self.assertTrue(results['OK'] > 0)
        self.assertEqual(results['FAILED'], 0)
        self.assertEqual(len(res), 1)

    @async_test
    async def test_failure(self, loop):

        @scenario(100)
        async def test_failing(session):
            raise ValueError()

        results = {'OK': 0, 'FAILED': 0}
        stream = asyncio.Queue()
        args = self.get_args()
        statsd = None

        await worker(1, loop, results, args, stream, statsd)

        self.assertEqual(results['OK'], 0)
        self.assertTrue(results['FAILED'] > 0)

    def test_shutdown(self):
        res = []

        @teardown()
        def _worker_teardown(num):
            res.append('BYE WORKER')

        @global_teardown()
        def _teardown():
            res.append('BYE')

        @scenario(100)
        async def test_two(session):
            os.kill(os.getpid(), signal.SIGTERM)

        args = self.get_args()
        results = runner(args)

        self.assertEqual(results['OK'], 1)
        self.assertEqual(results['FAILED'], 0)
        self.assertEqual(res, ['BYE WORKER', 'BYE'])



import random
import functools
import asyncio


_SCENARIO = []


def get_scenarios():
    return _SCENARIO


def _check_coroutine(func):
    if not asyncio.iscoroutinefunction(func):
        raise TypeError('%s needs to be a coroutine' % str(func))


def scenario(weight):
    def _scenario(func, *args, **kw):
        _check_coroutine(func)
        if weight > 0:
            _SCENARIO.append((weight, func, args, kw))

        @functools.wraps(func)
        def __scenario():
            return func(*args, **kw)
        return __scenario

    return _scenario


def pick_scenario():
    scenarios = get_scenarios()
    total = sum(item[0] for item in scenarios)
    selection = random.uniform(0, total)
    upto = 0
    for item in scenarios:
        weight = item[0]
        if upto + item[0] > selection:
            func, args, kw = item[1:]
            return func, args, kw
        upto += weight
    raise Exception('What')


_FIXTURES = {}


def get_fixture(name):
    return _FIXTURES.get(name)


def _fixture(name, coroutine=True):
    def __fixture(func, *args, **kw):
        if coroutine:
            _check_coroutine(func)
        if name in _FIXTURES:
            raise ValueError("You can't have two %r functions" % name)
        _FIXTURES[name] = func

        @functools.wraps(func)
        def ___fixture():
            return func(*args, **kw)

        return ___fixture
    return __fixture


def setup():
    return _fixture('setup')


def global_setup():
    return _fixture('global_setup', coroutine=False)


def teardown():
    return _fixture('teardown', coroutine=False)


def global_teardown():
    return _fixture('global_teardown', coroutine=False)



import json
import os
import sys
import argparse
import subprocess
import tempfile
import shutil
from collections import namedtuple

from molotov import __version__
from molotov.run import run


def clone_repo(github):
    # XXX security
    subprocess.check_call('git clone %s .' % github, shell=True)


def create_virtualenv(virtualenv, python):
    # XXX security
    subprocess.check_call('%s --python %s venv' % (virtualenv, python),
                          shell=True)


def install_reqs(reqfile):
    subprocess.check_call('./venv/bin/pip install -r %s' % reqfile,
                          shell=True)


_DEFAULTS = {'processes': False, 'verbose': False, 'scenario': 'loadtest.py',
             'workers': 1, 'duration': 10, 'quiet': False,
             'statsd': False, 'console': True, 'debug': False}


def run_test(**options):
    for option, value in _DEFAULTS.items():
        if option not in options:
            options[option] = value

    args = namedtuple('Arguments', options.keys())(**options)
    print('Running molotov with %s' % str(args))
    return run(args)


def main():
    parser = argparse.ArgumentParser(description='Github-based load test')

    parser.add_argument('--version', action='store_true', default=False,
                        help='Displays version and exits.')

    parser.add_argument('--virtualenv', type=str, default='virtualenv',
                        help='Virtualenv executable.')

    parser.add_argument('--python', type=str, default=sys.executable,
                        help='Python executable.')

    parser.add_argument('--config', type=str, default='molotov.json',
                        help='Path of the configuration file.')

    parser.add_argument('repo', help='Github repo', type=str)
    parser.add_argument('run', help='Test to run')

    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    tempdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tempdir)
    print('Working directory is %s' % tempdir)
    try:
        clone_repo(args.repo)
        config_file = os.path.join(tempdir, args.config)

        with open(config_file) as f:
            config = json.loads(f.read())

        # creating the virtualenv
        create_virtualenv(args.virtualenv, args.python)

        # install deps
        if 'requirements' in config['molotov']:
            install_reqs(config['molotov']['requirements'])

        # environment
        if 'env' in config['molotov']:
            for key, value in config['molotov']['env'].items():
                os.environ[key] = value

        run_test(**config['molotov']['tests'][args.run])
    except Exception:
        os.chdir(curdir)
        shutil.rmtree(tempdir, ignore_errors=True)
        raise


if __name__ == '__main__':
    main()



import gzip
import asyncio

from aiohttp.client_reqrep import ClientRequest
from yarl import URL

from molotov.session import LoggedClientSession
from molotov.tests.support import coserver, Response
from molotov.tests.support import TestLoop, async_test


class TestLoggedClientSession(TestLoop):

    @async_test
    async def test_encoding(self, loop):
        stream = asyncio.Queue()
        async with LoggedClientSession(loop, stream,
                                       verbose=True) as session:
            binary_body = b'MZ\x90\x00\x03\x00\x00\x00\x04\x00'
            response = Response(body=binary_body)
            await session.print_response(response)

        res = []
        while stream.qsize() > 0:
            line = await stream.get()
            res.append(line)

        wanted = "can't display this body"
        self.assertTrue(wanted in res[1])

    @async_test
    async def test_request(self, loop):
        with coserver():
            stream = asyncio.Queue()
            async with LoggedClientSession(loop, stream,
                                           verbose=True) as session:
                async with session.get('http://localhost:8888') as resp:
                    self.assertEqual(resp.status, 200)

            res = []
            while stream.qsize() > 0:
                line = await stream.get()
                res.append(line)

            self.assertTrue('GET http://127.0.0.1:8888' in res[1])

    @async_test
    async def test_gzipped_request(self, loop):
        stream = asyncio.Queue()
        async with LoggedClientSession(loop, stream,
                                       verbose=True) as session:
            binary_body = gzip.compress(b'some gzipped data')
            req = ClientRequest('GET', URL('http://example.com'),
                                data=binary_body)
            req.headers['Content-Encoding'] = 'gzip'
            await session.print_request(req)

        res = []
        while stream.qsize() > 0:
            line = await stream.get()
            res.append(line)

        self.assertTrue("Binary" in res[1], res)

    @async_test
    async def test_cantread_request(self, loop):
        stream = asyncio.Queue()
        async with LoggedClientSession(loop, stream,
                                       verbose=True) as session:
            binary_body = gzip.compress(b'some gzipped data')
            req = ClientRequest('GET', URL('http://example.com'),
                                data=binary_body)
            await session.print_request(req)

        res = []
        while stream.qsize() > 0:
            line = await stream.get()
            res.append(line)

        self.assertTrue("display this body" in res[1], res)



import sys
import shutil
import os
import argparse
from molotov import __version__


_DEFAULTS = {'target_dir': '.'}
_PREFIX = '> '
_HERE = os.path.dirname(__file__)


class ValidationError(Exception):
    pass


def _input(msg):
    return input(msg)


def _prompt(text, validator=None, default=None):
    while True:
        try:
            if default:
                res = _input(_PREFIX + '%s [%s]: ' % (text, default))
            else:
                res(_PREFIX + '%s: ' % text)

            if not res and default:
                res = default

            if validator:
                res = validator(res)

            return res
        except ValidationError as e:
            print(e)


def _yes(x):
    if x.upper() not in ('Y', 'YES', 'N', 'NO'):
        raise ValidationError("Please enter either 'y' or 'n'.")
    return x.upper() in ('Y', 'YES')


def _parser():
    parser = argparse.ArgumentParser(description='Quickstart')
    parser.add_argument('--version', action='store_true', default=False,
                        help='Displays version and exits.')

    return parser


def _copy_file(name, target_dir):
    print("…copying %r in %r" % (name, target_dir))
    target = os.path.join(target_dir, name)
    if os.path.exists(target):
        print("%r already exists. Cowardly stopping here" % target)
        sys.exit(1)
    shutil.copyfile(os.path.join(_HERE, name), target)


def main():
    parser = _parser()
    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    # XXX
    print('**** Molotov Quickstart ****')
    print('')
    print('Answer to a few questions to get started...')
    target_dir = _prompt("Target directory", default='.')
    create_makefile = _prompt("Create Makefile", default='y', validator=_yes)

    print('Generating Molotov test...')
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    if create_makefile:
        _copy_file('Makefile', target_dir)

    _copy_file('loadtest.py', target_dir)
    _copy_file('molotov.json', target_dir)

    print("")
    print("All done. Happy Breaking!")
    print("Go in %r" % target_dir)
    if create_makefile:
        print("Run 'make build' to get started...")


if __name__ == '__main__':
    main()



import tempfile
import unittest
import shutil
import os

from molotov import quickstart
from molotov.tests.support import set_args


class TestQuickStart(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _input(self, text):
        if 'Target directory' in text:
            return self.tempdir
        return 'y'

    def test_generate(self):
        quickstart._input = self._input

        with set_args('molostart'):
            quickstart.main()

        result = os.listdir(self.tempdir)
        result.sort()
        self.assertEqual(result, ['Makefile', 'loadtest.py', 'molotov.json'])

        # second runs stops
        with set_args('molostart'):
            try:
                quickstart.main()
                raise AssertionError()
            except SystemExit:
                pass



from molotov.api import pick_scenario, scenario, get_scenarios, setup
from molotov.tests.support import TestLoop


class TestUtil(TestLoop):
    def test_pick_scenario(self):

        @scenario(10)
        async def _one(self):
            pass

        @scenario(90)
        async def _two(self):
            pass

        picked = [pick_scenario()[0].__name__ for i in range(100)]
        ones = len([f for f in picked if f == '_one'])
        self.assertTrue(ones < 20)

    def test_no_scenario(self):
        @scenario(0)
        async def _one(self):
            pass

        @scenario(0)
        async def _two(self):
            pass

        self.assertEqual(get_scenarios(), [])

    def test_scenario_not_coroutine(self):
        try:
            @scenario(1)
            def _one(self):
                pass
        except TypeError:
            return
        raise AssertionError("Should raise")

    def test_setup_not_coroutine(self):
        try:
            @setup()
            def _setup(self):
                pass

            @scenario(90)
            async def _two(self):
                pass
        except TypeError:
            return
        raise AssertionError("Should raise")
